﻿using System.Xml.Serialization;

namespace NaGet.Packages
{
	/// <summary>
	/// インストーラの種類を表す。
	/// </summary>
	public enum InstallerType
	{
		/// <summary>
		/// インストーラ
		/// </summary>
		[XmlEnum(Name="installer")]
		EXEC_INSTALLER,
		/// <summary>
		/// Microsoft Software Installer
		/// </summary>
		[XmlEnum(Name="msi")]
		MSI_PACKAGE,
		/// <summary>
		/// 自己解凍書庫を含む書庫一般
		/// </summary>
		[XmlEnum(Name="archive")]
		ARCHIVE,
		/// <summary>
		/// インストールできない単なるデータ(BIOSなど)
		/// </summary>
		[XmlEnum(Name="cannotinstall")]
		CANNOT_INSTALL,
	}
	
	/// <summary>
	/// パッケージ情報を格納するクラス
	/// </summary>
	public class Package
	{
		/// <summary>
		/// パッケージ名
		/// </summary>
		public string Name;

		/// <summary>
		/// バージョン文字列
		/// </summary>
		public string Version;

		/// <summary>
		/// パッケージの概要
		/// </summary>
		public string Summary;
		
		/// <summary>
		/// パッケージの解説
		/// </summary>
		public string Description;

		/// <summary>
		/// 公式サイトのURL
		/// </summary>
		public LocationEntry Url;
		
		/// <summary>
		/// 製作者の名前
		/// </summary>
		public string Author;
		
		/// <summary>
		/// インストーラのタイプ
		/// </summary>
		public InstallerType Type;
		
				/// <summary>
		/// インストーラが書庫内に入っているかのフラグ
		/// </summary>
		public bool ArchivedInstaller = false;
		
		/// <summary>
		/// アンインストーラのレジストリのキー
		/// </summary>
		public string UninstallerKey;
		
		/// <summary>
		/// インストールスクリプト(MSBuildドキュメント)
		/// </summary>
		public string InstallScript;
		
		/// <summary>
		/// アーカイブインストーラのときのSystem32のコピー
		/// </summary>
		public string System32CopyFiles;
		
		/// <summary>
		/// インストーラのリソースの配列
		/// </summary>
		[XmlElement]
		public Installer[] Installer;

		/// <summary>
		/// 必要とされるパッケージの配列
		/// </summary>
		[XmlArray(IsNullable=true),XmlArrayItem("Entry")]
		public Entry[] Requires;
		
		/// <summary>
		/// 競合するパッケージの配列
		/// </summary>
		[XmlArray(IsNullable=true),XmlArrayItem("Entry")]
		public Entry[] Conflicts;
		
		/// <summary>
		/// タグ
		/// </summary>
		public string Tags;
		
		/// <summary>
		/// ライセンス
		/// </summary>
		public string License;
		
		/// <summary>
		/// 所属するパッケージリスト名称
		/// </summary>
		public string PackageListName;
	}

	/// <summary>
	/// パッケージ参照を示すクラス
	/// </summary>
	public class Entry
	{
		/// <summary>
		/// パッケージ名
		/// </summary>
		[XmlAttribute]
		public string Name;
		/// <summary>
		/// パッケージのバージョン比較の式
		/// </summary>
		[XmlAttribute]
		public string Flags;
		/// <summary>
		/// パッケージのバージョン
		/// </summary>
		public string Version;
	}

	public class Installer
	{
		/// <summary>
		/// コンストラクタ
		/// </summary>
		public Installer()
		{
		}
		
		public Platform Platform;
		
		/// <summary>
		/// ダウンロードリソースURLの配列
		/// </summary>
		public LocationEntry Url;
		
		/// <summary>
		/// ファイル検証用ハッシュ
		/// </summary>
		[XmlElement]
		public HashValue[] Hash;
	}
	
	/// <summary>
	/// リソースの位置を示すクラス
	/// </summary>
	public class LocationEntry
	{
		// TODO LocationEntryは果たして必要なのかの、考察
		
		/// <summary>
		/// コンストラクタ
		/// </summary>
		public LocationEntry()
		{
		}

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="sHref">リソースのURL</param>
		public LocationEntry(string sHref)
		{
			Href = sHref;
		}

		/// <summary>
		/// リソースのURL
		/// </summary>
		[XmlAttribute]
		public string Href;
	}
}
